/*
   Copyright (C) 2004,2005,2006,2007,2008,2009,2010 Roger While

   This file is part of OpenCOBOL.

   The OpenCOBOL compiler is free software: you can redistribute it
   and/or modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation, either version 3 of the
   License, or (at your option) any later version.

   OpenCOBOL is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with OpenCOBOL.  If not, see <http://www.gnu.org/licenses/>.
*/


#include	"config.h"
#include	"defaults.h"

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	"libcob.h"
#include	"tarstamp.h"

#define	CB_IMSG_SIZE	24
#define	CB_IVAL_SIZE	(80 - CB_IMSG_SIZE - 4)

static void COB_NOINLINE
print_version (void)
{
	int	year;
	int	day;
	char	buff[64];
	char	month[64];

	memset (buff, 0, sizeof(buff));
	memset (month, 0, sizeof(month));
	day = 0;
	year = 0;
	sscanf (__DATE__, "%s %d %d", month, &day, &year);
	if (day && year) {
		sprintf (buff, "%s %2.2d %4.4d %s", month, day, year, __TIME__);
	} else {
		sprintf (buff, "%s %s", __DATE__, __TIME__);
	}
	printf ("cobcrun (%s) %s.%d\n",
		PACKAGE_NAME, PACKAGE_VERSION, PATCH_LEVEL);
	puts ("Copyright (C) 2004,2005,2006,2007,2008,2009,2010 Roger While");
	printf ("Built    %s\nPackaged %s\n", buff, COB_TAR_DATE);
}

static void COB_NOINLINE
print_usage (void)
{
	printf ("Usage: cobcrun PROGRAM [param ...]\n");
	printf ("or   : cobcrun --help (-h)\n");
	printf ("       Print this help\n");
	printf ("or   : cobcrun --version (-V)\n");
	printf ("       Print version information\n");
	printf ("or   : cobcrun --info (-i)\n");
	printf ("       Print build information\n");
}

static void COB_NOINLINE
cobc_var_print (const char *msg, const char *val)
{
	char	*p;
	char	*token;
	size_t	n;
	size_t	toklen;

	printf ("%-*.*s : ", CB_IMSG_SIZE, CB_IMSG_SIZE, msg);
	if (strlen(val) <= CB_IVAL_SIZE) {
		printf("%s\n", val);
		return;
	}
	p = strdup (val);
	if (!p) {
		fprintf (stderr, "Memory allocation failure\n");
		return;
	}
	n = 0;
	token = strtok (p, " ");
	for (; token; token = strtok (NULL, " ")) {
		toklen = strlen (token) + 1;
		if ((n + toklen) > CB_IVAL_SIZE) {
			if (n) {
				printf ("\n%*.*s", CB_IMSG_SIZE + 3,
					CB_IMSG_SIZE + 3, " ");
			}
			n = 0;
		}
		printf ("%s%s", (n ? " " : ""), token);
		n += toklen;
	}
	printf ("\n");
	free (p);
}

static void COB_NOINLINE
print_info (void)
{
	char	buff[4];

	print_version ();
	printf ("Build information\n");
	cobc_var_print ("Build environment",	COB_BLD_BUILD);
	cobc_var_print ("CC",			COB_BLD_CC);
	cobc_var_print ("CPPFLAGS",		COB_BLD_CPPFLAGS);
	cobc_var_print ("CFLAGS",		COB_BLD_CFLAGS);
	cobc_var_print ("LD",			COB_BLD_LD);
	cobc_var_print ("LDFLAGS",		COB_BLD_LDFLAGS);
	printf ("\nOpenCOBOL information\n");
	cobc_var_print ("COB_CC",		COB_CC);
	cobc_var_print ("COB_CFLAGS",		COB_CFLAGS);
	cobc_var_print ("COB_LDFLAGS",		COB_LDFLAGS);
	cobc_var_print ("COB_LIBS",		COB_LIBS);
	cobc_var_print ("COB_CONFIG_DIR",	COB_CONFIG_DIR);
	cobc_var_print ("COB_COPY_DIR",		COB_COPY_DIR);
	cobc_var_print ("COB_LIBRARY_PATH",	COB_LIBRARY_PATH);
	cobc_var_print ("COB_MODULE_EXT",	COB_MODULE_EXT);
	cobc_var_print ("COB_EXEEXT",		COB_EXEEXT);

#if	defined(USE_LIBDL) || defined(_WIN32)
	cobc_var_print ("Dynamic loading",	"System");
#else
	cobc_var_print ("Dynamic loading",	"Libtool");
#endif

#ifdef	COB_PARAM_CHECK
	cobc_var_print ("\"CBL_\" param check",	"Enabled");
#else
	cobc_var_print ("\"CBL_\" param check",	"Disabled");
#endif

	sprintf (buff, "%d", WITH_VARSEQ);
	cobc_var_print ("Variable format",	buff);

#ifdef	COB_LI_IS_LL
	cobc_var_print ("BINARY-C-LONG",	"8 bytes");
#else
	cobc_var_print ("BINARY-C-LONG",	"4 bytes");
#endif

#ifdef	WITH_SEQRA_EXTFH
	cobc_var_print ("Sequential handler",	"External");
#else
	cobc_var_print ("Sequential handler",	"Internal");
#endif
#ifdef	WITH_INDEX_EXTFH
	cobc_var_print ("ISAM handler",		"External");
#endif
#ifdef	WITH_DB
	cobc_var_print ("ISAM handler",		"BDB");
#endif
#ifdef	WITH_CISAM
	cobc_var_print ("ISAM handler",		"C-ISAM (Experimental)");
#endif
#ifdef	WITH_DISAM
	cobc_var_print ("ISAM handler",		"D-ISAM (Experimental)");
#endif
#ifdef	WITH_VBISAM
	cobc_var_print ("ISAM handler",		"VBISAM (Experimental)");
#endif
}

int
main (int argc, char **argv)
{
	union cob_call_union	unifunc;

	if (unlikely (argc <= 1)) {
		print_usage ();
		return 1;
	}

	/* Quick check without getopt */
	if (unlikely(argv[1][0] == '-')) {
		if (!strncmp (argv[1], "--version", 10) ||
		    !strncmp (argv[1], "-V", 3)) {
			print_version ();
			return 0;
		}
		if (!strncmp (argv[1], "--help", 7) ||
		    !strncmp (argv[1], "-h", 3)) {
			print_usage ();
			return 0;
		}
		if (!strncmp (argv[1], "--info", 7) ||
		    !strncmp (argv[1], "-i", 3)) {
			print_info ();
			return 0;
		}
		fprintf (stderr, "Invalid option\n");
		return 1;
	}
	if (strlen (argv[1]) > 31) {
		fprintf (stderr, "Invalid PROGRAM name\n");
		return 1;
	}
	cob_init (argc - 1, &argv[1]);
	unifunc.func_void = cob_resolve (argv[1]);
	if (unifunc.func_void == NULL) {
		cob_call_error ();
	}
	cob_stop_run ( unifunc.funcint() );
}
